# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import TermTk as ttk

class SuperSimpleHorizontalLine(ttk.TTkWidget):
    def paintEvent(self, canvas):
        w,h = self.size()
        canvas.drawText(pos=(0,h-1), text='┕'+('━'*(w-2))+'┙',color=ttk.TTkColor.fg("#888888"))

class NotePad(ttk.TTkGridLayout):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        te = ttk.TTkTextEdit(readOnly=False, lineNumber=True)

        self.addItem(wrapLayout := ttk.TTkGridLayout(), 0,0)
        self.addItem(fontLayout := ttk.TTkGridLayout(columnMinWidth=1), 1,0)
        self.addWidget(te,2,0)

        wrapLayout.addWidget(ttk.TTkLabel(text="Wrap: ", maxWidth=6),0,0)
        wrapLayout.addWidget(lineWrap := ttk.TTkComboBox(list=['NoWrap','WidgetWidth','FixedWidth'], maxWidth=20),0,1)
        wrapLayout.addWidget(ttk.TTkLabel(text=" Type: ",maxWidth=7),0,2)
        wrapLayout.addWidget(wordWrap := ttk.TTkComboBox(list=['WordWrap','WrapAnywhere'], maxWidth=20, enabled=False),0,3)
        wrapLayout.addWidget(ttk.TTkLabel(text=" FixW: ",maxWidth=7),0,4)
        wrapLayout.addWidget(fixWidth := ttk.TTkSpinBox(value=te.wrapWidth(), maxWidth=5, maximum=500, minimum=10, enabled=False),0,5)
        wrapLayout.addWidget(ttk.TTkSpacer(),0,10)

        # Empty columns/cells are 1 char wide due to "columnMinWidth=1" parameter in the GridLayout
        #           1       3                    8                11
        #    0       2       4    5    6    7     9       10       12
        # 0  [ ] FG  [ ] BG  [ ] LineNumber
        # 1  ┌─────┐ ┌─────┐ ╒═══╕╒═══╕╒═══╕╒═══╕ ┌──────┐┌──────┐
        # 2  │     │ │     │ │ a ││ a ││ a ││ a │ │ UNDO ││ REDO │
        # 3  └─────┘ └─────┘ └───┘└───┘└───┘└───┘ ╘══════╛└──────┘ ┕━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━┙

        # Char Fg/Bg buttons
        fontLayout.addWidget(cb_fg := ttk.TTkCheckbox(text=" FG"),0,0)
        fontLayout.addWidget(btn_fgColor := ttk.TTkColorButtonPicker(border=True, enabled=False, maxSize=(7,3)),1,0)

        fontLayout.addWidget(cb_bg := ttk.TTkCheckbox(text=" BG"),0,2)
        fontLayout.addWidget(btn_bgColor := ttk.TTkColorButtonPicker(border=True, enabled=False, maxSize=(7   ,3)),1,2)

        fontLayout.addWidget(cb_linenumber := ttk.TTkCheckbox(text=" LineNumber", checked=True),0,4,1,3)

        # Char style buttons
        fontLayout.addWidget(btn_bold          := ttk.TTkButton(border=True, maxSize=(5,3), checkable=True, text=ttk.TTkString( 'a' , ttk.TTkColor.BOLD)        ),1,4)
        fontLayout.addWidget(btn_italic        := ttk.TTkButton(border=True, maxSize=(5,3), checkable=True, text=ttk.TTkString( 'a' , ttk.TTkColor.ITALIC)      ),1,5)
        fontLayout.addWidget(btn_underline     := ttk.TTkButton(border=True, maxSize=(5,3), checkable=True, text=ttk.TTkString(' a ', ttk.TTkColor.UNDERLINE)   ),1,6)
        fontLayout.addWidget(btn_strikethrough := ttk.TTkButton(border=True, maxSize=(5,3), checkable=True, text=ttk.TTkString(' a ', ttk.TTkColor.STRIKETROUGH)),1,7)

        # Undo/Redo buttons
        fontLayout.addWidget(btn_undo := ttk.TTkButton(border=True, maxSize=(8,3), enabled=te.isUndoAvailable(), text=' UNDO '),1,9 )
        fontLayout.addWidget(btn_redo := ttk.TTkButton(border=True, maxSize=(8,3), enabled=te.isRedoAvailable(), text=' REDO '),1,10)
        # Undo/Redo events
        te.undoAvailable.connect(btn_undo.setEnabled)
        te.redoAvailable.connect(btn_redo.setEnabled)
        btn_undo.clicked.connect(te.undo)
        btn_redo.clicked.connect(te.redo)

        # Useless custom horizontal bar for aestetic reason
        fontLayout.addWidget(SuperSimpleHorizontalLine(),0,12,2,1)

        @ttk.pyTTkSlot(ttk.TTkColor)
        def _currentColorChangedCB(format):
            if fg := format.foreground():
                cb_fg.setCheckState(ttk.TTkK.Checked)
                btn_fgColor.setEnabled()
                btn_fgColor.setColor(fg.invertFgBg())
            else:
                cb_fg.setCheckState(ttk.TTkK.Unchecked)
                btn_fgColor.setDisabled()

            if bg := format.background():
                cb_bg.setCheckState(ttk.TTkK.Checked)
                btn_bgColor.setEnabled()
                btn_bgColor.setColor(bg)
            else:
                cb_bg.setCheckState(ttk.TTkK.Unchecked)
                btn_bgColor.setDisabled()

            btn_bold.setChecked(format.bold())
            btn_italic.setChecked(format.italic())
            btn_underline.setChecked(format.underline())
            btn_strikethrough.setChecked(format.strikethrough())
            # ttk.TTkLog.debug(f"{fg=} {bg=} {bold=} {italic=} {underline=} {strikethrough=   }")

        te.currentColorChanged.connect(_currentColorChangedCB)

        def _setStyle():
            color = ttk.TTkColor()
            if cb_fg.checkState() == ttk.TTkK.Checked:
                color += btn_fgColor.color().invertFgBg()
            if cb_bg.checkState() == ttk.TTkK.Checked:
                color += btn_bgColor.color()
            if btn_bold.isChecked():
                color += ttk.TTkColor.BOLD
            if btn_italic.isChecked():
                color += ttk.TTkColor.ITALIC
            if btn_underline.isChecked():
                color += ttk.TTkColor.UNDERLINE
            if btn_strikethrough.isChecked():
                color += ttk.TTkColor.STRIKETROUGH
            cursor = te.textCursor()
            cursor.applyColor(color)
            cursor.setColor(color)
            te.setFocus()

        cb_fg.stateChanged.connect(lambda x: btn_fgColor.setEnabled(x==ttk.TTkK.Checked))
        cb_bg.stateChanged.connect(lambda x: btn_bgColor.setEnabled(x==ttk.TTkK.Checked))
        cb_fg.clicked.connect(lambda _: _setStyle())
        cb_bg.clicked.connect(lambda _: _setStyle())

        cb_linenumber.stateChanged.connect(lambda x: te.setLineNumber(x==ttk.TTkK.Checked))

        btn_fgColor.colorSelected.connect(lambda _: _setStyle())
        btn_bgColor.colorSelected.connect(lambda _: _setStyle())

        btn_bold.clicked.connect(_setStyle)
        btn_italic.clicked.connect(_setStyle)
        btn_underline.clicked.connect(_setStyle)
        btn_strikethrough.clicked.connect(_setStyle)

        lineWrap.setCurrentIndex(0)
        wordWrap.setCurrentIndex(1)

        fixWidth.valueChanged.connect(te.setWrapWidth)

        @ttk.pyTTkSlot(int)
        def _lineWrapCallback(index):
            if index == 0:
                te.setLineWrapMode(ttk.TTkK.NoWrap)
                wordWrap.setDisabled()
                fixWidth.setDisabled()
            elif index == 1:
                te.setLineWrapMode(ttk.TTkK.WidgetWidth)
                wordWrap.setEnabled()
                fixWidth.setDisabled()
            else:
                te.setLineWrapMode(ttk.TTkK.FixedWidth)
                te.setWrapWidth(fixWidth.value())
                wordWrap.setEnabled()
                fixWidth.setEnabled()

        lineWrap.currentIndexChanged.connect(_lineWrapCallback)

        @ttk.pyTTkSlot(int)
        def _wordWrapCallback(index):
            if index == 0:
                te.setWordWrapMode(ttk.TTkK.WordWrap)
            else:
                te.setWordWrapMode(ttk.TTkK.WrapAnywhere)

        wordWrap.currentIndexChanged.connect(_wordWrapCallback)





# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import TermTk as ttk

class SuperObject():
    @staticmethod
    def dumpParams(obj,exclude=[]):
        def _dumpPrimitive(val):
            return val
        def _dumpTTkString(val):
            return val.toAnsi(True)
        def _dumpTTkColor(val):
            return str(val)
        def _dumpTTkLayout(val):
            return type(val).__name__
        def _dumpFlag(val):
            return val
        def _dumpList(val, propType):
            ret = []
            for i,t in enumerate(propType):
                if t['type'] in (int,str,float,bool):
                    ret.append(_dumpPrimitive(val[i]))
                elif type(t['type']) in (list,tuple):
                    ttk.TTkLog.warn("Feature not Implemented yet")
                elif t['type'] is ttk.TTkLayout:
                    ret.append(_dumpTTkLayout(val[i]))
                elif t['type'] in (ttk.TTkString,'singleLineTTkString'):
                    ret.append(_dumpTTkString(val[i]))
                elif t['type'] is ttk.TTkColor:
                    ret.append(_dumpTTkColor(val[i]))
                elif t['type'] in ('singleFlag','multiFlag'):
                    ret.append(_dumpFlag(val[i]))
                else:
                    ttk.TTkLog.warn("Type not Recognised")
            return ret

        params = {}
        for cc in reversed(type(obj).__mro__):
            # if hasattr(cc,'_ttkProperties'):
            if issubclass(cc, ttk.TTkWidget) or issubclass(cc, ttk.TTkLayout):
                ccName = cc.__name__
                if ccName in ttk.TTkUiProperties:
                    for p in ttk.TTkUiProperties[ccName]['properties']:
                        if p in exclude: continue
                        prop = ttk.TTkUiProperties[ccName]['properties'][p]
                        propType = prop['get']['type']
                        propCb = prop['get']['cb']
                        # ttk.TTkLog.debug(ccName)
                        if propType in (int,str,float,bool):
                            params |= {p: _dumpPrimitive(propCb(obj))}
                        elif type(propType) in (list,tuple):
                            params |= {p: _dumpList(propCb(obj), propType)}
                        elif propType is ttk.TTkLayout:
                            params |= {p: _dumpTTkLayout(propCb(obj))}
                        elif propType in (ttk.TTkString,'singleLineTTkString'):
                            params |= {p: _dumpTTkString(propCb(obj))}
                        elif propType is ttk.TTkColor:
                            params |= {p: _dumpTTkColor(propCb(obj))}
                        elif propType in ('singleflag','multiflags'):
                            params |= {p: _dumpFlag(propCb(obj))}
                        else:
                            ttk.TTkLog.warn("Type not Recognised")
        return params



# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from random import randint

import TermTk as ttk

import ttkDesigner.app.superobj as so
from .superobj import SuperObject

class SuperWidgetContainer(so.SuperWidget):
    __slots__ = ('_superLayout')
    def __init__(self, designer, wid, *args, **kwargs):

        self._superLayout = so.SuperLayout(designer=designer, lay=wid.layout(),)
        self._superRootWidget = kwargs.get('superRootWidget',False)
        kwargs['layout'] = ttk.TTkGridLayout()
        kwargs['layout'].addWidget(self._superLayout)

        padt, padb, padl, padr = wid.getPadding()
        kwargs['paddingTop'] = padt
        kwargs['paddingBottom'] = padb
        kwargs['paddingLeft'] = padl
        kwargs['paddingRight'] = padr
        super().__init__(designer, wid, *args, **kwargs)

    def getSuperProperties(self):
        additions, exceptions, exclude = super().getSuperProperties()
        exceptions |= {
            'Layout' : {
                'get':  { 'cb': lambda _: self._superLayout.layout().__class__ , 'type':'singleflag',
                    'flags': {
                        'TTkLayout'     : ttk.TTkLayout     ,
                        'TTkGridLayout' : ttk.TTkGridLayout ,
                        'TTkVBoxLayout' : ttk.TTkVBoxLayout ,
                        'TTkHBoxLayout' : ttk.TTkHBoxLayout } },
                'set':  { 'cb': lambda _,l: self.changeSuperLayout(l) , 'type':'singleflag',
                    'flags': {
                        'TTkLayout'     : ttk.TTkLayout     ,
                        'TTkGridLayout' : ttk.TTkGridLayout ,
                        'TTkVBoxLayout' : ttk.TTkVBoxLayout ,
                        'TTkHBoxLayout' : ttk.TTkHBoxLayout } },
            }
        }
        return additions, exceptions, exclude

    def dumpDict(self):
        ret = super().dumpDict()
        ret |= {
            'layout': self._superLayout.dumpDict()
        }
        return ret

    @staticmethod
    def _swFromWidget(wid, swClass, *args, **kwargs):
        sw = swClass(wid=wid, *args, **kwargs)
        sw.changeSuperLayout(type(wid.layout()))
        return sw

    def dropEvent(self, evt) -> bool:
        padt, padb, padl, padr = self._wid.getPadding()
        # evt = evt.copy()
        evt.x-=padl
        evt.y-=padt
        return self._superLayout.dropEvent(evt)

    def changeSuperLayout(self, layout):
        sl = self._superLayout
        self.layout().removeWidget(sl)
        if layout == ttk.TTkVBoxLayout:
            sl = so.SuperLayoutVBox(designer=self._designer, lay=self._wid.layout(), selectable=False)
        elif layout == ttk.TTkHBoxLayout:
            sl = so.SuperLayoutHBox(designer=self._designer, lay=self._wid.layout(), selectable=False)
        elif layout == ttk.TTkGridLayout:
            sl = so.SuperLayoutGrid(designer=self._designer, lay=self._wid.layout(), selectable=False)
        else:
            sl = so.SuperLayout(    designer=self._designer, lay=self._wid.layout(), selectable=False)
        self._superLayout = sl
        self._wid.setLayout(layout())
        self.layout().addWidget(sl)
        self._designer.weModified.emit()

    def updateAll(self):
        self.setPadding(*(self._wid.getPadding()))
        super().updateAll()

    def paintEvent(self, canvas):
        w,h = self.size()
        if so.SuperWidget._showLayout:
            t,b,l,r = self._wid.getPadding()
            for y in range(h):
                canvas.drawText(pos=(0,y),text='',width=w,color=self._layoutColor)
            for y in range(t,h-b):
                canvas.drawText(pos=(l,y),text='',width=w-r-l,color=self._layoutPadColor)
            # canvas.fill(color=self._layoutColor)
            # canvas.fill(pos=(l,t), size=(w-r-l,h-b-t), color=self._layoutPadColor)
        else:
            self._wid.getCanvas().updateSize()
            self._wid.paintEvent(self._wid.getCanvas())
            canvas.paintCanvas(
                    self._wid.getCanvas(),
                    (    0,     0, w, h), # geometry
                    (    0,     0, w, h), # slice
                    (    0,     0, w, h)) # bound


# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import json

from TermTk import  TTkK, TTkLog, TTkColor, TTkHelper, TTkShortcut
from TermTk import TTkString
from TermTk import pyTTkSlot, pyTTkSignal

from TermTk import TTkWidget, TTkFrame, TTkButton, TTkLabel, TTkMenuButton
from TermTk import TTkTabWidget
from TermTk import TTkFileDialogPicker, TTkMessageBox

from TermTk import TTkGridLayout, TTkVBoxLayout, TTkHBoxLayout
from TermTk import TTkSplitter
from TermTk import TTkLogViewer, TTkKeyPressView
from TermTk import TTkUiLoader, TTkUtil

from .cfg  import *
from .about import *
from .widgetbox import WidgetBoxScrollArea
from .windoweditor import WindowEditor
from .treeinspector import TreeInspector
from .propertyeditor import PropertyEditor
from .signalsloteditor import SignalSlotEditor
from .quickexport import QuickExport
from .notepad import NotePad

from .superobj import SuperWidget, SuperWidgetFrame

import pickle

#
#      Mimic the QT Designer layout
#
#      ┌─────────────────────╥───────────────────────────────╥───────────────────┐
#      │                     ║┌─────────────────────────────┐║                   │
#      │                     ║│       ToolBar               │║  Tree Inspector   │
#      │                     ║├─────────────────────────────┤║                   │
#      │                     ║│                             │║                   │
#      │   Widget            ║│                             │║                   │
#      │   Box               ║│                             │╠═══════════════════╡
#      │                     ║│      Main Window            │║                   │
#      │                     ║│      Editor                 │║   Property        │
#      │                     ║│                             │║   Editor          │
#      │                     ║│                             │║                   │
#      │                     ║│                             │║                   │
#      │                     ║│                             │║                   │
#      │                     ║│                             │║                   │
#      │                     ║│                             │╠═══════════════════╡
#      │                     ║│                             │║                   │
#      │                     ║└─────────────────────────────┘║   Signal/Slot     │
#      │                     ╟───────────────────────────────╢                   │
#      │                     ║     LOG Viewer                ║   Editor          │
#      │                     ║                               ║                   │
#      └─────────────────────╨───────────────────────────────╨───────────────────┘
#

class TTkDesigner(TTkGridLayout):
    __slots__ = ('_main', '_toolBar', '_fileNameLabel', '_modified',
                 '_sigslotEditor', '_treeInspector', '_windowEditor', '_notepad',
                 '_fileName', '_currentPath',
                 '_snapshot', '_snapId', '_snapSaved',
                 # Signals
                 'weModified', 'thingSelected', 'widgetNameChanged'
                 )
    def __init__(self, fileName=None, *args, **kwargs):
        self._fileName = "untitled.tui.json"
        self._currentPath = self._currentPath =  os.path.abspath('.')

        self.weModified = pyTTkSignal()
        self.thingSelected = pyTTkSignal(TTkWidget, TTkWidget)
        self.widgetNameChanged = pyTTkSignal(str, str)

        self.resetSnapshot()

        super().__init__(*args, **kwargs)

        self.addWidget(mainSplit := TTkSplitter())
        mainSplit.addItem(widgetBoxLayout := TTkVBoxLayout())
        #mainSplit.addWidget(TTkButton(text='A',border=True))

        # mainSplit.addWidget(sa := TTkScrollArea())
        # sa.viewport().setLayout(TTkGridLayout())
        # sa.viewport().layout().addWidget(WindowEditor())

        self._notepad = NotePad()
        self._main = TTkVBoxLayout()
        self._toolBar = TTkHBoxLayout()
        self._windowEditor = WindowEditor(self)
        self._sigslotEditor = SignalSlotEditor(self)
        self._treeInspector = TreeInspector(self, self._windowEditor.viewport())
        self._fileNameLabel = TTkLabel(text=f"( {self._fileName} )", alignment=TTkK.CENTER_ALIGN)

        widgetBoxLayout.addWidget(topMenuFrame := TTkFrame(minHeight=1,maxHeight=1,border=False))
        widgetBoxLayout.addWidget(WidgetBoxScrollArea(self))

        self._main.addItem(self._toolBar)
        self._main.addWidget(self._windowEditor)

        mainSplit.addWidget(centralSplit := TTkSplitter(orientation=TTkK.VERTICAL))
        centralSplit.addWidget(self._main)
        centralSplit.addWidget(bottonTabWidget := TTkTabWidget(border=False))
        # centralSplit.addWidget(TTkLogViewer())
        bottonTabWidget.addTab(self._sigslotEditor,'Signal/Slot Editor')
        bottonTabWidget.addTab(TTkLogViewer(),'Logs')

        mainSplit.addWidget(rightSplit := TTkSplitter(orientation=TTkK.VERTICAL))

        rightSplit.addItem(self._treeInspector)
        rightSplit.addItem(propertyEditor := PropertyEditor(), title="Property Editor")
        # rightSplit.addItem(self._sigslotEditor)

        self.thingSelected.connect(lambda _,s : s.pushSuperControlWidget() if s.hasControlWidget() else None)
        self.thingSelected.connect(propertyEditor.setDetail)

        self.weModified.connect(self._treeInspector.refresh)
        self.weModified.connect(self._takeSnapshot)

        fileMenu = topMenuFrame.newMenubarTop().addMenu("&File")
        fileMenu.addMenu("&New").menuButtonClicked.connect(self.new)
        fileMenu.addMenu("&Open").menuButtonClicked.connect(self.open)
        fileMenu.addMenu("&Save").menuButtonClicked.connect(self.save)
        fileMenu.addMenu("Save &As...").menuButtonClicked.connect(self.saveAs)
        fileMenu.addSpacer()
        fileMenu.addMenu("&Import 🎁").menuButtonClicked.connect(self.importDictWin)
        fileMenu.addMenu("&Export 📦").menuButtonClicked.connect(self.quickExport)
        fileMenu.addSpacer()
        fileMenu.addMenu("E&xit").menuButtonClicked.connect(self.quit)

        extraMenu = topMenuFrame.newMenubarTop().addMenu("E&dit")
        extraMenu.addMenu("&Undo (CTRL+Z)").menuButtonClicked.connect(self.undo)
        extraMenu.addMenu("&Redo (CTRL+Y)").menuButtonClicked.connect(self.redo)
        extraMenu.addSpacer()
        extraMenu.addMenu("&Scratchpad 📝").menuButtonClicked.connect(self.scratchpad)
        extraMenu.addMenu("&KeypressView").menuButtonClicked.connect(self.keypressview)
        extraMenu.addSpacer()
        extraMenu.addMenu("&Preview...").menuButtonClicked.connect(self.preview)

        def _showAbout(btn):
            TTkHelper.overlay(None, About(), 30,10)
        def _showAboutTTk(btn):
            TTkHelper.overlay(None, TTkAbout(), 30,10)

        helpMenu = topMenuFrame.newMenubarTop().addMenu("&Help", alignment=TTkK.RIGHT_ALIGN)
        helpMenu.addMenu("About ...").menuButtonClicked.connect(_showAbout)
        helpMenu.addMenu("About ttk").menuButtonClicked.connect(_showAboutTTk)

        w,_ = self.size()
        mainSplit.setSizes([25,None,42])
        centralSplit.setSizes([None,8])

        self._toolBar.addWidget(btnPreview := TTkButton(maxWidth=12, text='Preview...'))
        self._toolBar.addWidget(btnExport := TTkButton(maxWidth=17, text='Quick Export 📦'))
        self._toolBar.addWidget(btnColors  := TTkButton(maxWidth=11, checkable=True, text=
                            TTkString("▣",TTkColor.fg("#ff0000")) +
                            TTkString("▣",TTkColor.fg("#ffff00")) +
                            TTkString("▣",TTkColor.fg("#00ff00")) +
                            TTkString("▣",TTkColor.fg("#00ffff")) +
                            TTkString("▣",TTkColor.fg("#0000ff")) + "🦄"))

        btnPreview.clicked.connect(self.preview)
        btnExport.clicked.connect(self.quickExport)
        btnColors.toggled.connect(self.toggleColors)

        self._toolBar.addWidget(self._fileNameLabel)

        TTkShortcut(TTkK.CTRL | TTkK.Key_Z).activated.connect(self.undo)
        TTkShortcut(TTkK.CTRL | TTkK.Key_Y).activated.connect(self.redo)

        if fileName:
            self._openFile(fileName)

    def modified(self):
        return self._modified

    def setModified(self, modified):
        # if self._modified == modified: return
        self._modified = modified
        if modified:
            self._fileNameLabel.setText(f"( ● {self._fileName} )")
        else:
            self._fileNameLabel.setText(f"( {self._fileName} )")
            self._snapSaved = self._snapId

    # Snapshot Logic:
    #   _snapshots = [ s0 , s1 , s2 , s3 , s4 , s5 , s6 ]
    #   _snapId = 3                   ^
    #
    # Undo:
    #   _snapId-- = 2            ^ = s2
    #   load s2
    #
    # Redo:
    #   _snapId++ = 4                      ^ = s4
    #   load s4
    #
    # Take Snapshot:
    #   Remove the forward Snapshots:
    #   _snapshots = [ s0 , s1 , s2 , s3 ]
    #   Append The new one and update snapId
    #   _snapshots = [ s0 , s1 , s2 , s3 , s4+ ]
    #   _snapId++ = 4

    @pyTTkSlot()
    def resetSnapshot(self):
        self._snapshot = []
        self._snapId = -1
        self._snapSaved = -1
        self._modified = False

    @pyTTkSlot()
    def _takeSnapshot(self):
        tui = self._windowEditor.dumpDict()
        connections = self._sigslotEditor.dumpDict()
        data = {
            'version':'2.0.0',
            'tui':tui,
            'connections':connections}
        TTkLog.debug(f"{len(pickle.dumps(data))=} {len(self._snapshot)=}")
        if ( self._snapshot and
             0 <= self._snapId < len(self._snapshot) and
             data == self._snapshot[self._snapId]): return
        self._snapshot = self._snapshot[:self._snapId+1]+[data]
        self._snapId = len(self._snapshot)-1
        self.setModified(True)

    def _loadSnapshot(self):
        self.weModified.disconnect(self._takeSnapshot)
        data = self._snapshot[self._snapId]
        sw = SuperWidget.loadDict(self, self._windowEditor.viewport(), data['tui'])
        self._windowEditor.importSuperWidget(sw)
        self._sigslotEditor.importConnections(data['connections'])
        self._treeInspector.refresh()
        self.weModified.connect(self._takeSnapshot)

    @pyTTkSlot()
    def undo(self):
        TTkLog.debug(f"Undo: {len(self._snapshot)=}")
        if not self._snapshot: return
        if self._snapId <= 0: return
        self._snapId -= 1
        self._loadSnapshot()

    @pyTTkSlot()
    def redo(self):
        TTkLog.debug(f"Undo: {len(self._snapshot)=}")
        if not self._snapshot: return
        if self._snapId >= len(self._snapshot)-1: return
        self._snapId += 1
        self._loadSnapshot()

    def getWidgets(self):
        widgets = set()
        def _getMenu(menu):
            widgets.add(menu)
            for item in menu._submenu:
                if issubclass(type(item), TTkMenuButton):
                    _getMenu(item)
        def _getItems(layoutItem):
            if layoutItem.layoutItemType() == TTkK.WidgetItem:
                superThing = layoutItem.widget()
                if issubclass(type(superThing), SuperWidget):
                    widgets.add(superThing._wid)
                for c in superThing.layout().children():
                    _getItems(c)
                # Chec and add all the menubuttons
                if issubclass(type(superThing), SuperWidgetFrame):
                    if(_mb := superThing._wid.menuBar(TTkK.TOP)):
                        if(_mbi := _mb._mbItems(TTkK.LEFT_ALIGN)):
                            for _ch in _mbi.children():
                                _getMenu(_ch.widget())
                        if(_mbi := _mb._mbItems(TTkK.CENTER_ALIGN)):
                            for _ch in _mbi.children():
                                _getMenu(_ch.widget())
                        if(_mbi := _mb._mbItems(TTkK.RIGHT_ALIGN)):
                            for _ch in _mbi.children():
                                _getMenu(_ch.widget())
                    if(_mb := superThing._wid.menuBar(TTkK.BOTTOM)):
                        if(_mbi := _mb._mbItems(TTkK.LEFT_ALIGN)):
                            for _ch in _mbi.children():
                                _getMenu(_ch.widget())
                        if(_mbi := _mb._mbItems(TTkK.CENTER_ALIGN)):
                            for _ch in _mbi.children():
                                _getMenu(_ch.widget())
                        if(_mbi := _mb._mbItems(TTkK.RIGHT_ALIGN)):
                            for _ch in _mbi.children():
                                _getMenu(_ch.widget())
        _getItems(self._windowEditor.getTTk().widgetItem())
        return widgets

    @pyTTkSlot(bool)
    def toggleColors(self, state):
        SuperWidget.toggleHighlightLayout.emit(state)

    @pyTTkSlot()
    def quickExport(self):
        tui = self._windowEditor.dumpDict()
        connections = self._sigslotEditor.dumpDict()
        data = {
            'version':'2.0.0',
            'tui':tui,
            'connections':connections}

        win = QuickExport(data)
        TTkHelper.overlay(None, win, 2, 2, modal=True)

    @pyTTkSlot()
    def scratchpad(self):
        win = TTkWindow(
                title="Mr Scratchpad 📝",
                size=(80,30),
                layout=self._notepad,
                flags=TTkK.WindowFlag.WindowMaximizeButtonHint|TTkK.WindowFlag.WindowCloseButtonHint)
        TTkHelper.overlay(None, win, 2, 2, toolWindow=True)

    @pyTTkSlot()
    def keypressview(self):
        win = TTkWindow(
                title="Mr Keypress 🔑🐁",
                size=(70,7),
                layout=(_l:=TTkGridLayout()),
                flags=TTkK.WindowFlag.WindowMaximizeButtonHint|TTkK.WindowFlag.WindowCloseButtonHint)
        _l.addWidget(TTkKeyPressView(maxHeight=3))
        TTkHelper.overlay(None, win, 2, 2, toolWindow=True)

    @pyTTkSlot()
    def preview(self):
        tui = self._windowEditor.dumpDict()
        connections = self._sigslotEditor.dumpDict()
        # for line in jj.split('\n'):
        #     TTkLog.debug(f"{line}")
        newUI = {
            'version':'2.0.0',
            'tui':tui,
            'connections':connections}
        jj =  json.dumps(newUI, indent=1)

        widget = TTkUiLoader.loadJson(jj)
        win = TTkWindow(
                title="Mr Terminal",
                size=(80,30),
                layout=TTkGridLayout(),
                flags=TTkK.WindowFlag.WindowMaximizeButtonHint|TTkK.WindowFlag.WindowCloseButtonHint)
        win.layout().addWidget(widget)
        TTkHelper.overlay(None, win, 2, 2, modal=True)

    @pyTTkSlot()
    def new(self, firstRun=False):
        def _newCB(cb=None):
            @pyTTkSlot()
            def _ret(cb=cb):
                def _newCB():
                    self.resetSnapshot()
                    self.weModified.emit()
                    self.setModified(False)
                    cb()
                if self.modified():
                    self.askToSave(TTkString( f'The current document has been modified, do you want to save it?\nIf you don\'t save, your changes will be lost.', TTkColor.BOLD),
                    cb=_newCB)
                else:
                    _newCB()
            return _ret

        newWindow = TTkUiLoader.loadFile(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../tui/newWindow.tui.json"))
        newWindow.getWidgetByName("BtnWindow"   ).clicked.connect(_newCB(cb=self._windowEditor.newWindow))
        newWindow.getWidgetByName("BtnContainer").clicked.connect(_newCB(cb=self._windowEditor.newContainer))
        newWindow.getWidgetByName("BtnFrame"    ).clicked.connect(_newCB(cb=self._windowEditor.newFrame))
        newWindow.getWidgetByName("BtnResFrame" ).clicked.connect(_newCB(cb=self._windowEditor.newResFrame))
        TTkHelper.overlay(self._windowEditor, newWindow, 10, 4, modal=True)

    @pyTTkSlot()
    def importDictWin(self):
        def _probeCompressedText(_text):
            import re
            ret = ""
            for _t in _text.split('\n'):
                if m := re.match(r'^ *["\']([A-Za-z0-9+/]+[=]{0,2})["\' +]*$',_t):
                    ret += m.group(1)
                elif not re.match(r'^ *$',_t): # exclude empty lines
                    return ""
            return ret
        newWindow = TTkUiLoader.loadFile(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../tui/quickImport.tui.json"))
        te = newWindow.getWidgetByName("TextEdit")
        def _importDict(te=te):
            text = te.toPlainText()
            if compressed := _probeCompressedText(text):
                dd = TTkUtil.base64_deflate_2_obj(compressed)
            else:
                try:
                    dd = eval(text)
                except Exception as e:
                    TTkLog.error(str(e))
                    messageBox = TTkMessageBox(text= str(e),icon=TTkMessageBox.Icon.Warning)
                    TTkHelper.overlay(None, messageBox, 5, 5, True)
                    return

            if type(dd) not in (str,dict):
                messageBox = TTkMessageBox(text= f"Input is {type(dd)}\nImport data must be a dict or \ncompressed String definition",icon=TTkMessageBox.Icon.Warning)
                TTkHelper.overlay(None, messageBox, 5, 5, True)
                return

            dd = TTkUiLoader.normalise(dd)
            sw = SuperWidget.loadDict(self, self._windowEditor.viewport(), dd['tui'])

            self._windowEditor.importSuperWidget(sw)
            self._sigslotEditor.importConnections(dd['connections'])
            self._treeInspector.refresh()
            newWindow.close()

        newWindow.getWidgetByName("BtnDict"   ).clicked.connect(_importDict)
        TTkHelper.overlay(self._windowEditor, newWindow, 10, 4, modal=True)

    def _updateFileName(self, fullPath):
        self._currentPath =  os.path.dirname(os.path.abspath(fullPath))
        self._fileName =  os.path.basename(os.path.abspath(fullPath))
        self._fileNameLabel.setText(f"( {self._fileName} )")

    def _openFile(self, fileName):
        TTkLog.info(f"Open: {fileName}")
        self._updateFileName(fileName)

        with open(fileName) as fp:
            dd = json.load(fp)
            dd = TTkUiLoader.normalise(dd)
            sw = SuperWidget.loadDict(self, self._windowEditor.viewport(), dd['tui'])
            self._windowEditor.importSuperWidget(sw)
            self._sigslotEditor.importConnections(dd['connections'])
            self._treeInspector.refresh()

    @pyTTkSlot()
    def open(self):
        filePicker = TTkFileDialogPicker(pos = (3,3), size=(80,30), caption="Open", path=self._currentPath, fileMode=TTkK.FileMode.ExistingFile ,filter="TTk Tui Files (*.tui.json);;Json Files (*.json);;All Files (*)")
        filePicker.pathPicked.connect(self._openFile)
        TTkHelper.overlay(None, filePicker, 5, 5, True)

    @pyTTkSlot()
    def save(self):
        return self._saveToFile(os.path.join(self._currentPath,self._fileName))

    def _saveToFile(self, fileName):
        TTkLog.info(f"Saving to: {fileName}")
        self._updateFileName(fileName)

        tui = self._windowEditor.dumpDict()
        connections = self._sigslotEditor.dumpDict()
        newUI = {
            'version':'2.0.0',
            'tui':tui,
            'connections':connections}
        jj =  json.dumps(newUI, indent=1)

        with open(fileName,'w') as fp:
            fp.write(jj)

    @pyTTkSlot()
    def saveAs(self, cb=None):
        def _approveFile(fileName):
            if os.path.exists(fileName):
                @pyTTkSlot(TTkMessageBox.StandardButton)
                def _cb(btn):
                    if btn == TTkMessageBox.StandardButton.Save:
                        self._saveToFile(fileName)
                    elif btn == TTkMessageBox.StandardButton.Cancel:
                        return
                    if cb:
                        cb()
                messageBox = TTkMessageBox(
                    text= (
                        TTkString( f'A file named "{os.path.basename(fileName)}" already exists.\nDo you want to replace it?', TTkColor.BOLD) +
                        TTkString( f'\n\nReplacing it will overwrite its contents.') ),
                    icon=TTkMessageBox.Icon.Warning,
                    standardButtons=TTkMessageBox.StandardButton.Discard|TTkMessageBox.StandardButton.Save|TTkMessageBox.StandardButton.Cancel)
                messageBox.buttonSelected.connect(_cb)
                TTkHelper.overlay(None, messageBox, 5, 5, True)
            else:
                self._saveToFile(fileName)
        filePicker = TTkFileDialogPicker(pos = (3,3), size=(80,30), acceptMode=TTkK.AcceptMode.AcceptSave, caption="Save As...", path=os.path.join(self._currentPath,self._fileName), fileMode=TTkK.FileMode.AnyFile ,filter="TTk Tui Files (*.tui.json);;Json Files (*.json);;All Files (*)")
        filePicker.pathPicked.connect(_approveFile)
        TTkHelper.overlay(None, filePicker, 5, 5, True)

    @pyTTkSlot()
    def askToSave(self, text="", cb=None):
        messageBox = TTkMessageBox(
            text=text,
            icon=TTkMessageBox.Icon.Warning,
            standardButtons=TTkMessageBox.StandardButton.Discard|TTkMessageBox.StandardButton.Save|TTkMessageBox.StandardButton.Cancel)
        @pyTTkSlot(TTkMessageBox.StandardButton)
        def _cb(btn):
            if btn == TTkMessageBox.StandardButton.Save:
                self.saveAs(cb=cb)
            elif btn == TTkMessageBox.StandardButton.Cancel:
                return
            elif cb:
                cb()
            messageBox.buttonSelected.clear()
        messageBox.buttonSelected.connect(_cb)
        TTkHelper.overlay(None, messageBox, 5, 5, True)

    def quit(self):
        if self.modified():
            self.askToSave(
                TTkString( f'Do you want to save the changes to this document before closing?\nIf you don\'t save, your changes will be lost.', TTkColor.BOLD),
                cb=TTkHelper.quit)
        else:
            TTkHelper.quit()



# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import TermTk as ttk

def QuickExport(data):
    newWindow = ttk.TTkUiLoader.loadFile(os.path.join(os.path.dirname(os.path.abspath(__file__)),"../tui/quickExport.tui.json"))
    te = newWindow.getWidgetByName("TextEdit")

    ttk.pyTTkSlot()
    def _useNormmal():
        te.setLineWrapMode(ttk.TTkK.WidgetWidth)
        te.setWordWrapMode(ttk.TTkK.WordWrap)
        te.setText('from TermTk import TTkUtil, TTkUiLoader, TTk\n')
        te.append(f'# Data generated using ttkDesigner')
        te.append('widget = TTkUiLoader.loadDict(')
        te.append(str(data)+ ')')
        te.append('\nroot=TTk()\nroot.layout().addWidget(widget)\nroot.mainloop()\n')

    ttk.pyTTkSlot()
    def _useCompressed():
        te.setLineWrapMode(ttk.TTkK.NoWrap)
        b64str = ttk.TTkUtil.obj_inflate_2_base64(data)
        te.setText('from TermTk import TTkUtil, TTkUiLoader, TTk\n')
        te.append(f'# Data generated using ttkDesigner')
        te.append('widget = TTkUiLoader.loadDict(TTkUtil.base64_deflate_2_obj(')
        b64list = '    "' + '" +\n    "'.join([b64str[i:i+128] for i in range(0,len(b64str),128)]) + '"))'
        te.append(b64list)
        te.append('\nroot=TTk()\nroot.layout().addWidget(widget)\nroot.mainloop()\n')

    def _saveToFile(fileName):
        ttk.TTkLog.info(f"Saving to: {fileName}")
        with open(fileName,'w') as fp:
             fp.write(te.toPlainText())
        newWindow.close()

    ttk.pyTTkSlot(str)
    def _checkSaveFile(fileName):
        if os.path.exists(fileName):
            messageBox = ttk.TTkMessageBox(
                text= (
                    ttk.TTkString( f'A file named "{os.path.basename(fileName)}" already exists.\nDo you want to replace it?', ttk.TTkColor.BOLD) +
                    ttk.TTkString( f'\n\nReplacing it will overwrite its contents.') ),
                icon=ttk.TTkMessageBox.Icon.Warning,
                standardButtons=ttk.TTkMessageBox.StandardButton.Discard|ttk.TTkMessageBox.StandardButton.Save|ttk.TTkMessageBox.StandardButton.Cancel)
            messageBox.buttonSelected.connect(lambda btn : _saveToFile(fileName) if btn == ttk.TTkMessageBox.StandardButton.Save else None)
            ttk.TTkHelper.overlay(None, messageBox, 5, 5, True)
        else:
            _saveToFile(fileName)

    newWindow.getWidgetByName("BtnNormal").clicked.connect(_useNormmal)
    newWindow.getWidgetByName("BtnCompressed").clicked.connect(_useCompressed)
    newWindow.getWidgetByName("BtnSave").filePicked.connect(_checkSaveFile)
    _useCompressed()

    return newWindow



# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import TermTk as ttk
import ttkDesigner.app.superobj as so
from .superlayout import SuperLayout

class _SuperExpandButton(ttk.TTkButton):
    NONE     = 0x00
    PRESSED  = 0x01
    RELEASED = 0x02
    __slots__ = ('superExpandButtonDragged','superExpandButtonHidden','_mouseStatus')
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._direction = ttk.TTkK.TOP
        # This signal is required by the superGridLayout to identify a dragging
        # that may change the padding of the linked widget
        self.superExpandButtonDragged = ttk.pyTTkSignal(int,int,bool)
        self.superExpandButtonHidden = ttk.pyTTkSignal()
        self._mouseStatus = self.NONE

    def setDirection(self, d):
        self._direction = d

    def _processInput(self, kevt, mevt):
        if not mevt:
            self.hide()
            return
        ax,ay = ttk.TTkHelper.absPos(self)
        w,h = self.size()
        # x,y,w,h = self._boundaries
        mx,my = mevt.x, mevt.y
        if ( self._mouseStatus == self.NONE and
             ( not (0 <= (mx-ax) < w) or
               not (0 <= (my-ay) < h) ) ) :
            self.hide()

    def mousePressEvent(self, evt):
        self._mouseStatus = self.PRESSED
        return super().mousePressEvent(evt)
    def mouseReleaseEvent(self, evt):
        self._mouseStatus = self.NONE
        return super().mouseReleaseEvent(evt)

    def mouseDragEvent(self, evt) -> bool:
        # ttk.TTkLog.debug(f"Drag - {evt}")
        x,y,w,h = self.geometry()
        self.superExpandButtonDragged.emit(evt.x+x, evt.y+y, 0<=evt.x<w and 0<=evt.y<h)
        return True

    def show(self):
        ttk.TTkInput.inputEvent.connect(self._processInput)
        return super().show()

    def hide(self):
        self.superExpandButtonHidden.emit()
        ttk.TTkInput.inputEvent.disconnect(self._processInput)
        return super().hide()

    def paintEvent(self, canvas):
        # '▶','◀','▼','▲'
        w,h = self.size()
        if w==1:
            canvas.drawText(text='╽', pos=(0,0),   color=ttk.TTkColor.fg("FFFF00"))
            canvas.drawText(text='╿', pos=(0,h-1), color=ttk.TTkColor.fg("FFFF00"))
            for yy in range(1,h-1):
                canvas.drawText(text='┃', pos=(0, yy), color=ttk.TTkColor.fg("FFFF00"))
        elif h==1:
            txt = '╼'+'━'*(w-2)+'╾'
            canvas.drawText(text=txt, pos=(0,0), width=w, color=ttk.TTkColor.fg("FFFF00"))

        ch = {
            ttk.TTkK.TOP    : '▲',
            ttk.TTkK.BOTTOM : '▼',
            ttk.TTkK.LEFT   : '◀',
            ttk.TTkK.RIGHT  : '▶',
            ttk.TTkK.HORIZONTAL : '↔',
            ttk.TTkK.VERTICAL   : '↕',
        }.get(self._direction, 'X')

        x,y = 0,0
        if w==1 and h>4:
            y = h//2-2
            h = 4
        elif h==1 and w>4:
            x = w//2-2
            w = 4
        canvas.fill(pos=(x,y), size=(w,h), char=ch, color=ttk.TTkColor.fg("#FF0000")+ttk.TTkColor.bg("#FFFF44"))

class SuperLayoutGrid(SuperLayout):
    __slots__ = ('_expandButton','_dragOver','_spanOver','_expandStuff','_orientation','_snappId')
    def __init__(self, *args, **kwargs):
        kwargs['layout'] = ttk.TTkGridLayout()
        super().__init__(*args, **kwargs)
        self._expandButton = _SuperExpandButton()
        self.rootLayout().addWidget(self._expandButton)
        self._expandButton.hide()
        self._expandButton.clicked.connect(self._clickExpand)
        self._expandButton.superExpandButtonDragged.connect(self._expandButtonDragged)
        self._expandButton.superExpandButtonHidden.connect(self._expandButtonHidden)
        # self._expandButton.superExpandButtonDragged.connect()
        self._dragOver = None
        self._spanOver = None
        self._snappId  = None
        self._expandStuff = None
        self._orientation = ttk.TTkK.HORIZONTAL|ttk.TTkK.VERTICAL

    def dragEnterEvent(self, evt) -> bool:
        # ttk.TTkLog.debug(f"Enter")
        _, __, ___, self._dragOver = self._processDragOver(evt.x,evt.y)
        return True
    def dragLeaveEvent(self, evt) -> bool:
        # ttk.TTkLog.debug(f"Leave")
        self._dragOver = None
        self.update()
        return True
    def dragMoveEvent(self, evt) -> bool:
        # ttk.TTkLog.debug(f"Move")
        _, __, ___, self._dragOver = self._processDragOver(evt.x,evt.y)
        return True
    def dropEvent(self, evt) -> bool:
        self._dragOver = None
        self._pushRow, self._pushCol, self._direction, self._dragOver = self._processDragOver(evt.x,evt.y)
        return super().dropEvent(evt)

    @ttk.pyTTkSlot()
    def _expandButtonHidden(self):
        self._spanOver = None
        self.update()

    def removeSuperWidget(self, sw):
        super().removeSuperWidget(sw)
        self.layout().repack()

    def mouseMoveEvent(self, evt) -> bool:
        # ttk.TTkLog.debug(f"Move {evt}")
        dir, pos, wid, spanOver = self._processMouseOver(evt.x, evt.y)

        if not wid or not dir:
            self._expandButton.hide()
            return super().mouseMoveEvent(evt)
        x,y,w,h = wid.geometry()
        ebs = {
            ttk.TTkK.TOP    : (x,     y,     w, 1),
            ttk.TTkK.BOTTOM : (x,     y+h-1, w, 1),
            ttk.TTkK.LEFT   : (x,     y,     1, h),
            ttk.TTkK.RIGHT  : (x+w-1, y,     1, h),
        }.get(pos, None)

        if not ebs:
            self._expandButton.hide()
        else:
            self._expandButton.setGeometry(*ebs)
            self._expandButton.setDirection(dir)
            self._expandButton.raiseWidget(raiseParent=False)
            self._expandButton.show()
        self._expandStuff = (dir,wid)
        self._spanOver = spanOver
        return True

    def addSuperWidget(self, sw):
        self._dragOver = None
        if self._direction == ttk.TTkK.HORIZONTAL or self._orientation == ttk.TTkK.HORIZONTAL:
            self.layout().insertColumn(self._pushCol)
        elif self._direction == ttk.TTkK.VERTICAL or self._orientation == ttk.TTkK.VERTICAL:
            self.layout().insertRow(self._pushRow)
        self.layout().addWidget(sw, self._pushRow, self._pushCol,1,1)

    @ttk.pyTTkSlot(int, int)
    def _expandButtonDragged(self, dx, dy, expand):
        # ttk.TTkLog.debug(f"Drag - {(dx,dy)}")
        self._snappId = None
        self.update()
        if expand:
            self._snappId = self._expandButton
        elif self._spanOver:
            for snapId in self._spanOver:
                ((x,y,w,h),_) = snapId
                ttk.TTkLog.debug(f"{snapId=} {(x,dx,w)=} {(y,dy,h)=}")
                if x<=dx<x+w and y<=dy<y+h:
                    # ttk.TTkLog.debug(f"XXXX -> {snapId=}")
                    self._snappId = snapId
                    return

    ttk.pyTTkSlot()
    def _clickExpand(self):
        if not self._expandButton.isVisible(): return
        self._expandButton.hide()
        dir, wid = self._expandStuff
        row = wid._row
        col = wid._col
        rowspan = wid._rowspan
        colspan = wid._colspan
        self.layout().removeItem(wid)
        if ttk.TTkK.TOP and row==0:
            self.layout().insertRow(0)
            row+=1
        elif ttk.TTkK.LEFT and col==0:
            self.layout().insertColumn(0)
            col+=1
        rc = (row,col,rowspan,colspan)
        if self._snappId == self._expandButton:
            rc = {
                ttk.TTkK.TOP    : (row-1,col,  rowspan+1,colspan),
                ttk.TTkK.BOTTOM : (row,  col,  rowspan+1,colspan),
                ttk.TTkK.LEFT   : (row,  col-1,rowspan,colspan+1),
                ttk.TTkK.RIGHT  : (row,  col,  rowspan,colspan+1),
            }.get(dir, (row,col,rowspan,colspan))
        elif self._snappId:
            (_,rc) = self._snappId
        self._snappId = None
        self.layout().addItem(wid,*rc)

    def _processMouseOver(self, x, y):
        # cehck the closest edge
        col, row, dir, pos, placesSpan = 0, 0, None, None, []
        wid = None

        if type(self.layout()) != ttk.TTkGridLayout:
            return dir,pos,wid,placesSpan

        # Retrieve a list of widths,heights
        rows,cols = self.layout().gridSize()
        if not rows or not cols: return dir,pos,wid,placesSpan

        horSizes, verSizes = self.layout().getSizes()

        # Find the row/col where the pointer is in
        for col,(a,b) in enumerate(horSizes):
            if a <= x < a+b: break
        for row,(a,b) in enumerate(verSizes):
            if a <= y < a+b: break

        # ix, iw = horSizes[col]
        # iy, ih = verSizes[row]

        wid = self.layout().itemAtPosition(row,col)
        if wid == None:
            return dir,pos,wid,placesSpan

        col = wid._col
        row = wid._row
        rowspan = wid._rowspan
        colspan = wid._colspan

        widX,widY,widW,widH = wid.geometry()
        widMaxW,widMaxH = wid.maximumSize()

        #Top
        if ( y==widY ):
            placesSpan = [[[widX,iy,widW,1],[row+i,col,rowspan-i,colspan]] for i,(iy,ih) in enumerate(verSizes[row+1:row+rowspan],1)]
            pos = ttk.TTkK.TOP if placesSpan else None
            dir = ttk.TTkK.VERTICAL
            if ( widMaxH>widH and
                 not any([self.layout().itemAtPosition(row-1,col+cs) for cs in range(colspan)])):
                dir = pos = ttk.TTkK.TOP
        #Bottom
        elif (widY+widH==y+1):
            placesSpan = [[[widX,iy+ih-1,widW,1],[row,col,i,colspan]] for i,(iy,ih) in enumerate(verSizes[row:row+rowspan-1],1)]
            pos = ttk.TTkK.BOTTOM if placesSpan else None
            dir = ttk.TTkK.VERTICAL
            if ( widMaxH>widH and
                 not any([self.layout().itemAtPosition(row+rowspan,col+cs) for cs in range(colspan)])):
                dir = pos = ttk.TTkK.BOTTOM
        #Left
        elif (x==widX):
            placesSpan = [[[ix,widY,1,widH],[row,col+i,rowspan,colspan-i]] for i,(ix,iw) in enumerate(horSizes[col+1:col+colspan],1)]
            pos = ttk.TTkK.LEFT if placesSpan else None
            dir = ttk.TTkK.HORIZONTAL
            if ( widMaxW>widW and
                 not any([self.layout().itemAtPosition(row+rs,col-1) for rs in range(rowspan)])):
                dir = pos = ttk.TTkK.LEFT
        #Right
        elif (widX+widW==x+1):
            placesSpan = [[[ix+iw-1,widY,1,widH],[row,col,rowspan,i]] for i,(ix,iw) in enumerate(horSizes[col:col+colspan-1],1)]
            pos = ttk.TTkK.RIGHT if placesSpan else None
            dir = ttk.TTkK.HORIZONTAL
            if ( widMaxW>widW and
                 not any([self.layout().itemAtPosition(row+rs,col+colspan) for rs in range(rowspan)])):
                dir = pos = ttk.TTkK.RIGHT
        else:
            wid=None

        self._snappId=self._expandButton

        # ttk.TTkLog.debug(f"Move {dir} {wid}")

        # ttk.TTkLog.debug(f"{horSizes=}")
        # ttk.TTkLog.debug(f"{verSizes=}")
        # ttk.TTkLog.debug(f"{row=} {col=} {dir=} {self._dragOver=}")
        self.update()
        return dir,pos,wid,placesSpan

    def _processDragOver(self, x, y):
        # cehck the closest edge
        col, row, dir = 0,0,None
        ret = None

        # Retrieve a list of widths,heights
        rows,cols = self.layout().gridSize()
        if not rows or not cols: return col,row,dir,ret

        horSizes, verSizes = self.layout().getSizes()

        # Find the row/col where the pointer is in
        for col,(a,b) in enumerate(horSizes):
            if a <= x < a+b: break
        for row,(a,b) in enumerate(verSizes):
            if a <= y < a+b: break

        ix, iw = horSizes[col]
        iy, ih = verSizes[row]

        dt = y-iy
        db = iy+ih-y-1
        dl = x-ix
        dr = ix+iw-x-1
        dmin = min(dt,db,dl,dr)

        if self.layout().itemAtPosition(row,col) == None:
            ret = (ix, iy, iw, ih)
        else:
            #Top - we are closer to this edge
            if ((dt==dmin) and (self._orientation & ttk.TTkK.VERTICAL) and
                (  row==0 or
                 ( row>0 and self.layout().itemAtPosition(row,col) != self.layout().itemAtPosition(row-1,col)))):
                dir = ttk.TTkK.VERTICAL
                if row>0 and self.layout().itemAtPosition(row-1,col):
                    ret = (ix,    iy-1,    iw, 2)
                else:
                    ret = (ix,    iy,    iw, 1)
            #Bottom - we are closer to this edge
            if ((db==dmin) and (self._orientation & ttk.TTkK.VERTICAL) and
                self.layout().itemAtPosition(row,col) != self.layout().itemAtPosition(row+1,col)):
                dir = ttk.TTkK.VERTICAL
                if row<rows-1 and self.layout().itemAtPosition(row+1,col):
                    ret = (ix,    iy+ih-1, iw, 2)
                else:
                    ret = (ix,    iy+ih-1, iw, 1)
            #Left - we are closer to this edge
            if ((dl==dmin) and (self._orientation & ttk.TTkK.HORIZONTAL) and
                (  col==0 or
                 ( col>0 and self.layout().itemAtPosition(row,col) != self.layout().itemAtPosition(row,col-1)))):
                dir = ttk.TTkK.HORIZONTAL
                if col>0 and self.layout().itemAtPosition(row,col-1):
                    ret = (ix-1,  iy,    2, ih)
                else:
                    ret = (ix,    iy,    1, ih)
            #Right - we are closer to this edge
            if ((dr==dmin) and (self._orientation & ttk.TTkK.HORIZONTAL) and
                self.layout().itemAtPosition(row,col) != self.layout().itemAtPosition(row,col+1)):
                dir = ttk.TTkK.HORIZONTAL
                if col<cols-1 and self.layout().itemAtPosition(row,col+1):
                    ret = (ix+iw-1, iy,    2, ih)
                else:
                    ret = (ix+iw-1, iy,    1, ih)

            # If we are on the edge of the item push to the next spot
            if dir == ttk.TTkK.HORIZONTAL and ix+iw==x+1:
                col+=1
            if dir == ttk.TTkK.VERTICAL   and iy+ih==y+1:
                row+=1

        # ttk.TTkLog.debug(f"{horSizes=}")
        # ttk.TTkLog.debug(f"{verSizes=}")
        # ttk.TTkLog.debug(f"{row=} {col=} {dir=} {self._dragOver=}")
        self.update()
        return row, col, dir, ret

    # Stupid hack to paint on top of the child widgets
    def paintChildCanvas(self):
        super().paintChildCanvas()

        canvas = self.getCanvas()
        def _lineDraw(x,y,w,h,color):
            if h==1 and w==1:
                canvas.drawText(text='◉', pos=(x,y), color=color)
            elif w==1:
                canvas.drawText(text='╽', pos=(x,y),     color=color)
                canvas.drawText(text='╿', pos=(x,y+h-1), color=color)
                for yy in range(y+1,y+h-1):
                    canvas.drawText(text='┃', pos=(x, yy), color=color)
            elif h==1:
                txt = '╼'+'━'*(w-2)+'╾'
                canvas.drawText(text=txt, pos=(x,y), width=w, color=color)
            else:
                canvas.drawBox(pos=(x,y), size=(w,h), color=color)

        if self._dragOver is not None:
            _lineDraw(*self._dragOver, ttk.TTkColor.fg("FFFF00"))
        if self._spanOver:
            for (geom,_) in self._spanOver:
                _lineDraw(*geom, ttk.TTkColor.fg("FFFF00"))
        if self._snappId and self._snappId != self._expandButton:
            (geom,_) = self._snappId
            _lineDraw(*geom, ttk.TTkColor.bg("88FF88")+ttk.TTkColor.fg("#000044"))





# MIT License
#
# Copyright (c) 2023 Eugenio Parodi <ceccopierangiolieugenio AT googlemail DOT com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from TermTk.TTkCore.log import TTkLog
from TermTk.TTkCore.color import TTkColor
from TermTk.TTkCore.string import TTkString
from TermTk import TTkAbout, TTkWindow
from .cfg import TTkDesignerCfg

class About(TTkAbout):
    # designerTxt = [
    #     "   ____          _                 ",
    #     "╭─|    \ ___ ___|_|___ ___ ___ ___ ",
    #     "│ |  |  | -_|_ -| | . |   | -_|  _|",
    #     "TTk____/|___|___|_|_  |_|_|___|_|  ",
    #     "╰─────────────────|___|──────────┄ "]
    designerTxt = [
                                                                        TTkString("   ____          _                 ",TTkColor.fg("#FFFFFF")),
        TTkString("╭─", TTkColor.fg("#4400FF"))+                        TTkString(  "|    \ ___ ___|_|___ ___ ___ ___ ",TTkColor.fg("#F8FFE8")),
        TTkString("│ ", TTkColor.fg("#5500FF"))+                        TTkString(  "|  |  | -_|_ -| | . |   | -_|  _|",TTkColor.fg("#EFFFCF")),
        TTkString("TTk",TTkColor.fg("#00FF66")+TTkColor.bg("#7700FF"))+ TTkString(   "____/|___|___|_|_  |_|_|___|_|  ",TTkColor.fg("#E8FFB8")),
        TTkString("╰─────────────────",TTkColor.fg("#9900FF"))+         TTkString(                  "|___|",            TTkColor.fg("#DFFF9F"))+TTkString("──────────┄ ",TTkColor.fg("#BB00FF"))]

    __slots__=('_image')
    def __init__(self, *args, **kwargs):
        TTkAbout.__init__(self, *args, **kwargs)

        self.setTitle('[PierCecco Cecco] Eugenio Parodi proudly presents...')
        self.resize(56,15)

    def paintEvent(self, canvas):
        c = [0xFF,0xFF,0xFF]
        for y, line in enumerate(About.designerTxt):
            canvas.drawText(pos=(13,3+y),text=line)
        canvas.drawText(pos=(20, 9),text=f"  Version: {TTkDesignerCfg.version}", color=TTkColor.fg('#AAAAFF'))
        canvas.drawText(pos=(14,11),text=f"Powered By, pyTermTk")
        canvas.drawText(pos=( 2,13),text=f"https://github.com/ceccopierangiolieugenio/pyTermTk", color=TTkColor.fg('#44FFFF'))

        TTkWindow.paintEvent(self, canvas)
